import mock

from datetime import datetime

from django.test import TestCase
from keycloak.openid_connect import KeycloakOpenidConnect
from keycloak.authz import KeycloakAuthz

from django_keycloak.factories import OpenIdConnectProfileFactory
from django_keycloak.tests.mixins import MockTestCaseMixin

import django_keycloak.services.oidc_profile


class ServicesKeycloakOpenIDProfileGetActiveAccessTokenTestCase(
        MockTestCaseMixin, TestCase):

    def setUp(self):
        self.mocked_get_active_access_token = self.setup_mock(
            'django_keycloak.services.oidc_profile'
            '.get_active_access_token'
        )

        self.oidc_profile = OpenIdConnectProfileFactory(
            access_token='access-token',
            expires_before=datetime(2018, 3, 5, 1, 0, 0),
            refresh_token='refresh-token'
        )
        self.oidc_profile.realm.client.openid_api_client = mock.MagicMock(
            spec_set=KeycloakOpenidConnect)
        self.oidc_profile.realm.client.authz_api_client = mock.MagicMock(
            spec_set=KeycloakAuthz)
        self.oidc_profile.realm.client.authz_api_client.entitlement\
            .return_value = {
                'rpt': 'RPT_VALUE'
            }
        self.oidc_profile.realm.certs = {'cert': 'cert-value'}

    def test(self):
        django_keycloak.services.oidc_profile.get_entitlement(
            oidc_profile=self.oidc_profile
        )
        self.oidc_profile.realm.client.authz_api_client.entitlement\
            .assert_called_once_with(
                token=self.mocked_get_active_access_token.return_value
            )
        self.oidc_profile.realm.client.openid_api_client.decode_token\
            .assert_called_once_with(
                token='RPT_VALUE',
                key=self.oidc_profile.realm.certs,
                options={
                    'verify_signature': True,
                    'exp': True,
                    'iat': True,
                    'aud': True
                }
            )



import logging

from django.contrib.auth.models import Permission
from requests.exceptions import HTTPError

import django_keycloak.services.client

logger = logging.getLogger(__name__)


def synchronize(client):
    """
    :param django_keycloak.models.Client client:
    :return:
    """
    keycloak_client_id = django_keycloak.services.client.get_keycloak_id(
        client=client)

    role_api = client.admin_api_client.realms.by_name(client.realm.name)\
        .clients.by_id(keycloak_client_id).roles

    for permission in Permission.objects.all():

        try:
            role_api.create(name=permission.codename,
                            description=permission.name)
        except HTTPError as e:
            if e.response.status_code != 409:
                raise

        else:
            continue

        # Update role
        role_api.by_name(permission.codename) \
            .update(name=permission.codename, description=permission.name)



import json
import logging
import uuid

from django.contrib.auth.models import Permission
from django.db import models
from django.conf import settings
from django.utils import timezone
from django.utils.functional import cached_property

logger = logging.getLogger(__name__)


class Server(models.Model):

    url = models.CharField(max_length=255)

    internal_url = models.CharField(
        max_length=255,
        blank=True,
        null=True,
        help_text='URL on internal netwerk calls. For example when used with '
                  'Docker Compose. Only supply when internal calls should go '
                  'to a different url as the end-user will communicate with.'
    )

    def __str__(self):
        return self.url


class Realm(models.Model):

    server = models.ForeignKey(Server, related_name='realms',
                               on_delete=models.CASCADE)

    name = models.CharField(max_length=255, unique=True,
                            help_text='Name as known on the Keycloak server. '
                                      'This name is used in the API paths '
                                      'of this Realm.')
    _certs = models.TextField()

    @property
    def certs(self):
        return json.loads(self._certs)

    @certs.setter
    def certs(self, content):
        self._certs = json.dumps(content)

    _well_known_oidc = models.TextField(blank=True)

    @property
    def well_known_oidc(self):
        return json.loads(self._well_known_oidc)

    @well_known_oidc.setter
    def well_known_oidc(self, content):
        self._well_known_oidc = json.dumps(content)

    _keycloak_realm = None

    @cached_property
    def realm_api_client(self):
        """
        :rtype: keycloak.realm.Realm
        """
        if self._keycloak_realm is None:
            import django_keycloak.services.realm
            self._keycloak_realm = django_keycloak.services.realm.\
                get_realm_api_client(realm=self)
        return self._keycloak_realm

    def __str__(self):
        return self.name


class Client(models.Model):

    realm = models.OneToOneField(Realm, related_name='client',
                                 on_delete=models.CASCADE)

    client_id = models.CharField(max_length=255)
    secret = models.CharField(max_length=255)

    service_account_profile = models.OneToOneField(
        settings.KEYCLOAK_OIDC_PROFILE_MODEL,
        on_delete=models.CASCADE,
        null=True
    )

    @cached_property
    def admin_api_client(self):
        """
        :rtype: keycloak.admin.KeycloakAdmin
        """
        import django_keycloak.services.client
        return django_keycloak.services.client.get_admin_client(client=self)

    @cached_property
    def openid_api_client(self):
        """
        :rtype: keycloak.openid_connect.KeycloakOpenidConnect
        """
        import django_keycloak.services.client
        return django_keycloak.services.client.get_openid_client(client=self)

    @cached_property
    def authz_api_client(self):
        """
        :rtype: keycloak.authz.KeycloakAuthz
        """
        import django_keycloak.services.client
        return django_keycloak.services.client.get_authz_api_client(
            client=self)

    @cached_property
    def uma1_api_client(self):
        """
        :rtype: keycloak.uma1.KeycloakUMA1
        """
        import django_keycloak.services.client
        return django_keycloak.services.client.get_uma1_client(client=self)

    def __str__(self):
        return self.client_id


class Role(models.Model):

    client = models.ForeignKey('django_keycloak.Client',
                               related_name='roles',
                               on_delete=models.CASCADE)
    permission = models.ForeignKey(Permission,
                                   on_delete=models.CASCADE)
    reference = models.CharField(max_length=50)

    class Meta(object):
        unique_together = (
            ('client', 'permission')
        )


class TokenModelAbstract(models.Model):

    access_token = models.TextField(null=True)
    expires_before = models.DateTimeField(null=True)

    refresh_token = models.TextField(null=True)
    refresh_expires_before = models.DateTimeField(null=True)

    class Meta(object):
        abstract = True

    @property
    def is_active(self):
        if not self.access_token or not self.expires_before:
            return False

        return self.expires_before > timezone.now()


class OpenIdConnectProfileAbstract(TokenModelAbstract):

    sub = models.CharField(max_length=255, unique=True)

    realm = models.ForeignKey('django_keycloak.Realm',
                              related_name='openid_profiles',
                              on_delete=models.CASCADE)

    class Meta(object):
        abstract = True

    @property
    def jwt(self):
        """
        :rtype: dict
        """
        if not self.is_active:
            return None
        client = self.realm.client
        return client.openid_api_client.decode_token(
            token=self.access_token,
            key=client.realm.certs,
            algorithms=client.openid_api_client.well_known[
                'id_token_signing_alg_values_supported']
        )


class RemoteUserOpenIdConnectProfile(OpenIdConnectProfileAbstract):

    is_remote = True
    _user = None

    class Meta(OpenIdConnectProfileAbstract.Meta):
        swappable = 'KEYCLOAK_OIDC_PROFILE_MODEL'

    def get_user(self):
        if self._user is None:
            import django_keycloak.services.oidc_profile
            self._user = django_keycloak.services.oidc_profile. \
                get_remote_user_from_profile(
                    oidc_profile=self
                )
        return self._user

    def set_user(self, user):
        import django_keycloak.services.oidc_profile
        RemoteUserModel = django_keycloak.services.oidc_profile\
            .get_remote_user_model()
        if not isinstance(user, RemoteUserModel):
            raise RuntimeError('Can\'t set a non-remote user to the {}'.format(
                type(self)))

        self._user = user

    user = property(get_user, set_user)


class OpenIdConnectProfile(OpenIdConnectProfileAbstract):

    is_remote = False

    user = models.OneToOneField(settings.AUTH_USER_MODEL,
                                related_name='oidc_profile',
                                on_delete=models.CASCADE)

    class Meta(RemoteUserOpenIdConnectProfile.Meta):
        swappable = 'KEYCLOAK_OIDC_PROFILE_MODEL'


class Nonce(models.Model):

    state = models.UUIDField(default=uuid.uuid4, unique=True)
    redirect_uri = models.CharField(max_length=255)
    next_path = models.CharField(max_length=255, null=True)


class ExchangedToken(TokenModelAbstract):

    oidc_profile = models.ForeignKey(
        settings.KEYCLOAK_OIDC_PROFILE_MODEL,
        on_delete=models.CASCADE
    )
    remote_client = models.ForeignKey('django_keycloak.RemoteClient',
                                      related_name='exchanged_tokens',
                                      on_delete=models.CASCADE)

    class Meta(object):
        unique_together = (
            ('oidc_profile', 'remote_client'),
        )


class RemoteClient(models.Model):

    name = models.CharField(max_length=255)
    realm = models.ForeignKey('django_keycloak.Realm',
                              related_name='remote_clients',
                              on_delete=models.CASCADE)



from django.contrib import admin

from django_keycloak.admin.realm import RealmAdmin
from django_keycloak.admin.server import ServerAdmin
from django_keycloak.models import Server, Realm

admin.site.register(Realm, RealmAdmin)
admin.site.register(Server, ServerAdmin)



from keycloak.realm import KeycloakRealm

try:
    from urllib.parse import urlparse
except ImportError:
    from urlparse import urlparse


def get_realm_api_client(realm):
    """
    :param django_keycloak.models.Realm realm:
    :return keycloak.realm.Realm:
    """
    headers = {}
    server_url = realm.server.url
    if realm.server.internal_url:
        # An internal URL is configured. We add some additional settings to let
        # Keycloak think that we access it using the server_url.
        server_url = realm.server.internal_url
        parsed_url = urlparse(realm.server.url)
        headers['Host'] = parsed_url.netloc

        if parsed_url.scheme == 'https':
            headers['X-Forwarded-Proto'] = 'https'

    return KeycloakRealm(server_url=server_url, realm_name=realm.name,
                         headers=headers)


def refresh_certs(realm):
    """
    :param django_keycloak.models.Realm realm:
    :rtype django_keycloak.models.Realm
    """
    realm.certs = realm.client.openid_api_client.certs()
    realm.save(update_fields=['_certs'])
    return realm


def refresh_well_known_oidc(realm):
    """
    Refresh Open ID Connect .well-known

    :param django_keycloak.models.Realm realm:
    :rtype django_keycloak.models.Realm
    """
    server_url = realm.server.internal_url or realm.server.url

    # While fetching the well_known we should not use the prepared URL
    openid_api_client = KeycloakRealm(
        server_url=server_url,
        realm_name=realm.name
    ).open_id_connect(client_id='', client_secret='')

    realm.well_known_oidc = openid_api_client.well_known.contents
    realm.save(update_fields=['_well_known_oidc'])
    return realm


def get_issuer(realm):
    """
    Get correct issuer to validate the JWT against. If an internal URL is
    configured for the server it will be replaced with the public one.

    :param django_keycloak.models.Realm realm:
    :return: issuer
    :rtype: str
    """
    issuer = realm.well_known_oidc['issuer']
    if realm.server.internal_url:
        return issuer.replace(realm.server.internal_url, realm.server.url, 1)
    return issuer



# Configure the model which need to be used to store the Open ID connect
# profile. There are two choices:
# - django_keycloak.OpenIdConnectProfile (Default) a local User object get
#   created for the logged in identity.
# - django_keycloak.RemoteUserOpenIdConnectProfile with this model there will
#   be no local user stored for the logged in identity.
KEYCLOAK_OIDC_PROFILE_MODEL = 'django_keycloak.OpenIdConnectProfile'

# Class which will be used as User object in case of the remote user OIDC
# Profile
KEYCLOAK_REMOTE_USER_MODEL = 'django_keycloak.remote_user.KeycloakRemoteUser'
KEYCLOAK_PERMISSIONS_METHOD = 'role'  # 'role' of 'resource'



from django.test import TestCase, override_settings

from django_keycloak.factories import OpenIdConnectProfileFactory
from django_keycloak.tests.mixins import MockTestCaseMixin
from django_keycloak.auth.backends import KeycloakAuthorizationBase


@override_settings(KEYCLOAK_PERMISSIONS_METHOD='resource')
class BackendsKeycloakAuthorizationBaseGetKeycloakPermissionsTestCase(
        MockTestCaseMixin, TestCase):

    def setUp(self):
        self.backend = KeycloakAuthorizationBase()

        self.profile = OpenIdConnectProfileFactory()

        self.setup_mock(
            'django_keycloak.services.oidc_profile.get_entitlement',
            return_value={
                'authorization': {
                    'permissions': [
                        {
                            'resource_set_name': 'Resource',
                            'scopes': [
                                'Read',
                                'Update'
                            ]
                        },
                        {
                            'resource_set_name': 'Resource2'
                        }
                    ]
                }
            })

    def test_get_keycloak_permissions(self):
        """
        Case: The permissions are requested from Keycloak, which are returned
        by get_entitlement as a decoded RPT.
        Expect: The permissions are extracted from the RPT and are returned
        in a list.
        """
        permissions = self.backend.get_keycloak_permissions(
            user_obj=self.profile.user)

        self.assertListEqual(
            ['Read_Resource', 'Update_Resource', 'Resource2'],
            permissions
        )



import mock

from datetime import datetime

from django.contrib.auth import get_user_model
from django.test import TestCase
from freezegun import freeze_time
from keycloak.openid_connect import KeycloakOpenidConnect

from django_keycloak.factories import ClientFactory
from django_keycloak.models import OpenIdConnectProfile
from django_keycloak.tests.mixins import MockTestCaseMixin

import django_keycloak.services.oidc_profile


class ServicesKeycloakOpenIDProfileUpdateOrCreateTestCase(MockTestCaseMixin,
                                                          TestCase):

    def setUp(self):
        self.client = ClientFactory(
            realm___certs='{}',
            realm___well_known_oidc='{"issuer": "https://issuer"}'
        )
        self.client.openid_api_client = mock.MagicMock(
            spec_set=KeycloakOpenidConnect)
        self.client.openid_api_client.authorization_code.return_value = {
            'id_token': 'id-token',
            'expires_in': 600,
            'refresh_expires_in': 3600,
            'access_token': 'access-token',
            'refresh_token': 'refresh-token'
        }
        self.client.openid_api_client.well_known = {
            'id_token_signing_alg_values_supported': ['signing-alg']
        }
        self.client.openid_api_client.decode_token.return_value = {
            'sub': 'some-sub',
            'email': 'test@example.com',
            'given_name': 'Some given name',
            'family_name': 'Some family name'
        }

    @freeze_time('2018-03-01 00:00:00')
    def test_create(self):
        django_keycloak.services.oidc_profile.update_or_create_from_code(
            client=self.client,
            code='some-code',
            redirect_uri='https://redirect'
        )
        self.client.openid_api_client.authorization_code\
            .assert_called_once_with(code='some-code',
                                     redirect_uri='https://redirect')
        self.client.openid_api_client.decode_token.assert_called_once_with(
            token='id-token',
            key=dict(),
            algorithms=['signing-alg'],
            issuer='https://issuer'
        )

        profile = OpenIdConnectProfile.objects.get(sub='some-sub')
        self.assertEqual(profile.sub, 'some-sub'),
        self.assertEqual(profile.access_token, 'access-token')
        self.assertEqual(profile.refresh_token, 'refresh-token')
        self.assertEqual(profile.expires_before, datetime(
            year=2018, month=3, day=1, hour=0, minute=10, second=0
        ))
        self.assertEqual(profile.refresh_expires_before, datetime(
            year=2018, month=3, day=1, hour=1, minute=0, second=0
        ))

        user = profile.user
        self.assertEqual(user.username, 'some-sub')
        self.assertEqual(user.first_name, 'Some given name')
        self.assertEqual(user.last_name, 'Some family name')

    @freeze_time('2018-03-01 00:00:00')
    def test_update(self):
        UserModel = get_user_model()
        user = UserModel.objects.create(
            username='some-sub',
            email='',
            first_name='',
            last_name=''
        )
        profile = OpenIdConnectProfile.objects.create(
            realm=self.client.realm,
            sub='some-sub',
            user=user,
            access_token='another-access-token',
            expires_before=datetime.now(),
            refresh_token='another-refresh-token',
            refresh_expires_before=datetime.now()
        )

        django_keycloak.services.oidc_profile.update_or_create_from_code(
            client=self.client,
            code='some-code',
            redirect_uri='https://redirect'
        )
        self.client.openid_api_client.authorization_code\
            .assert_called_once_with(code='some-code',
                                     redirect_uri='https://redirect')
        self.client.openid_api_client.decode_token.assert_called_once_with(
            token='id-token',
            key=dict(),
            algorithms=['signing-alg'],
            issuer='https://issuer'
        )

        profile.refresh_from_db()
        self.assertEqual(profile.sub, 'some-sub')
        self.assertEqual(profile.access_token, 'access-token')
        self.assertEqual(profile.refresh_token, 'refresh-token')
        self.assertEqual(profile.expires_before, datetime(
            year=2018, month=3, day=1, hour=0, minute=10, second=0
        ))
        self.assertEqual(profile.refresh_expires_before, datetime(
            year=2018, month=3, day=1, hour=1, minute=0, second=0
        ))

        user = profile.user
        user.refresh_from_db()
        self.assertEqual(user.username, 'some-sub')
        self.assertEqual(user.first_name, 'Some given name')
        self.assertEqual(user.last_name, 'Some family name')



from django.test import TestCase

from django_keycloak.factories import ServerFactory, RealmFactory
from django_keycloak.tests.mixins import MockTestCaseMixin

import django_keycloak.services.realm


class ServicesRealmGetRealmApiClientTestCase(
        MockTestCaseMixin, TestCase):

    def setUp(self):
        self.server = ServerFactory(
            url='https://some-url',
            internal_url=''
        )

        self.realm = RealmFactory(
            server=self.server,
            name='test-realm'
        )

    def test_get_realm_api_client(self):
        """
        Case: a realm api client is requested for a realm on a server without
        internal_url.
        Expected: a KeycloakRealm client is returned with settings based on the
        provided realm. The server_url in the client is the provided url.
        """
        client = django_keycloak.services.realm.\
            get_realm_api_client(realm=self.realm)

        self.assertEqual(client.server_url, self.server.url)
        self.assertEqual(client.realm_name, self.realm.name)

    def test_get_realm_api_client_with_internal_url(self):
        """
        Case: a realm api client is requested for a realm on a server with
        internal_url.
        Expected: a KeycloakRealm client is returned with settings based on the
        provided realm. The server_url in the client is the provided url.
        """
        self.server.internal_url = 'https://some-internal-url'

        client = django_keycloak.services.realm.\
            get_realm_api_client(realm=self.realm)

        self.assertEqual(client.server_url, self.server.internal_url)
        self.assertEqual(client.realm_name, self.realm.name)



# Generated by Django 2.0.3 on 2018-03-22 20:59

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('django_keycloak', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='ExchangedToken',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True,
                                        serialize=False, verbose_name='ID')),
                ('access_token', models.TextField(null=True)),
                ('expires_before', models.DateTimeField(null=True)),
                ('refresh_token', models.TextField(null=True)),
                ('refresh_expires_before', models.DateTimeField(null=True)),
                ('oidc_profile', models.ForeignKey(
                    on_delete=django.db.models.deletion.CASCADE,
                    to='django_keycloak.OpenIdConnectProfile')),
            ],
        ),
        migrations.CreateModel(
            name='RemoteClient',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True,
                                        serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255)),
                ('realm', models.ForeignKey(
                    on_delete=django.db.models.deletion.CASCADE,
                    related_name='remote_clients',
                    to='django_keycloak.Realm'
                )),
            ],
        ),
        migrations.AddField(
            model_name='exchangedtoken',
            name='remote_client',
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name='exchanged_tokens',
                to='django_keycloak.RemoteClient'
            ),
        ),
        migrations.AlterUniqueTogether(
            name='exchangedtoken',
            unique_together={('oidc_profile', 'remote_client')},
        ),
    ]
